@@ -42,6 +42,23 @@
 #define DEFAULT_INPUT_REQUEST_TIMEOUT 120 * 1000
 #define DEFAULT_BROWSER_LAUNCH_TIMEOUT 300 * 1000
 
+#define MAINFILE "main.conf"
+#define CONFIGMAINFILE CONFIGDIR "/" MAINFILE
+
+static char *default_auto_connect[] = {
+	"wifi",
+	"ethernet",
+	"cellular",
+	NULL
+};
+
+static char *default_blacklist[] = {
+	"vmnet",
+	"vboxnet",
+	"virbr",
+	NULL
+};
+
 static struct {
 	connman_bool_t bg_scan;
 	char **pref_timeservers;
@@ -50,6 +67,9 @@ static struct {
 	char **fallback_nameservers;
 	unsigned int timeout_inputreq;
 	unsigned int timeout_browserlaunch;
+	char **blacklisted_interfaces;
+	connman_bool_t allow_hostname_updates;
+	connman_bool_t single_tech;
 } connman_settings  = {
 	.bg_scan = TRUE,
 	.pref_timeservers = NULL,
@@ -58,6 +78,34 @@ static struct {
 	.fallback_nameservers = NULL,
 	.timeout_inputreq = DEFAULT_INPUT_REQUEST_TIMEOUT,
 	.timeout_browserlaunch = DEFAULT_BROWSER_LAUNCH_TIMEOUT,
+	.blacklisted_interfaces = NULL,
+	.allow_hostname_updates = TRUE,
+	.single_tech = FALSE,
+};
+
+#define CONF_BG_SCAN                    "BackgroundScanning"
+#define CONF_PREF_TIMESERVERS           "FallbackTimeservers"
+#define CONF_AUTO_CONNECT               "DefaultAutoConnectTechnologies"
+#define CONF_PREFERRED_TECHS            "PreferredTechnologies"
+#define CONF_FALLBACK_NAMESERVERS       "FallbackNameservers"
+#define CONF_TIMEOUT_INPUTREQ           "InputRequestTimeout"
+#define CONF_TIMEOUT_BROWSERLAUNCH      "BrowserLaunchTimeout"
+#define CONF_BLACKLISTED_INTERFACES     "NetworkInterfaceBlacklist"
+#define CONF_ALLOW_HOSTNAME_UPDATES     "AllowHostnameUpdates"
+#define CONF_SINGLE_TECH                "SingleConnectedTechnology"
+
+static const char *supported_options[] = {
+	CONF_BG_SCAN,
+	CONF_PREF_TIMESERVERS,
+	CONF_AUTO_CONNECT,
+	CONF_PREFERRED_TECHS,
+	CONF_FALLBACK_NAMESERVERS,
+	CONF_TIMEOUT_INPUTREQ,
+	CONF_TIMEOUT_BROWSERLAUNCH,
+	CONF_BLACKLISTED_INTERFACES,
+	CONF_ALLOW_HOSTNAME_UPDATES,
+	CONF_SINGLE_TECH,
+	NULL
 };
 
 static GKeyFile *load_config(const char *file)
@@ -113,8 +161,6 @@ static char **parse_fallback_nameservers(char **nameservers, gsize len)
 {
 	char **servers;
 	int i, j;
-	struct addrinfo hints;
-	struct addrinfo *addr;
 
 	servers = g_try_new0(char *, len + 1);
 	if (servers == NULL)
@@ -123,60 +169,91 @@ static char **parse_fallback_nameservers(char **nameservers, gsize len)
 	i = 0;
 	j = 0;
 	while (nameservers[i] != NULL) {
-		memset(&hints, 0, sizeof(struct addrinfo));
-		hints.ai_flags = AI_NUMERICHOST;
-		addr = NULL;
-		if (getaddrinfo(nameservers[i], NULL, &hints, &addr) == 0) {
+		if (connman_inet_check_ipaddress(nameservers[i]) > 0) {
 			servers[j] = g_strdup(nameservers[i]);
 			j += 1;
 		}
-
-		freeaddrinfo(addr);
 		i += 1;
 	}
 
 	return servers;
 }
 
+static void check_config(GKeyFile *config)
+{
+	char **keys;
+	int j;
+
+	if (config == NULL)
+		return;
+
+	keys = g_key_file_get_groups(config, NULL);
+
+	for (j = 0; keys != NULL && keys[j] != NULL; j++) {
+		if (g_strcmp0(keys[j], "General") != 0)
+			connman_warn("Unknown group %s in %s",
+						keys[j], MAINFILE);
+	}
+
+	g_strfreev(keys);
+
+	keys = g_key_file_get_keys(config, "General", NULL, NULL);
+
+	for (j = 0; keys != NULL && keys[j] != NULL; j++) {
+		connman_bool_t found;
+		int i;
+
+		found = FALSE;
+		for (i = 0; supported_options[i] != NULL; i++) {
+			if (g_strcmp0(keys[j], supported_options[i]) == 0) {
+				found = TRUE;
+				break;
+			}
+		}
+		if (found == FALSE && supported_options[i] == NULL)
+			connman_warn("Unknown option %s in %s",
+						keys[j], MAINFILE);
+	}
+
+	g_strfreev(keys);
+}
+
 static void parse_config(GKeyFile *config)
 {
 	GError *error = NULL;
 	gboolean boolean;
 	char **timeservers;
+	char **interfaces;
 	char **str_list;
 	gsize len;
-	char *default_auto_connect[] = {
-		"wifi",
-		"ethernet",
-		"cellular",
-		NULL
-	};
 	int timeout;
 
 	if (config == NULL) {
 		connman_settings.auto_connect =
 			parse_service_types(default_auto_connect, 3);
+		connman_settings.blacklisted_interfaces =
+			g_strdupv(default_blacklist);
 		return;
 	}
 
-	DBG("parsing main.conf");
+	DBG("parsing %s", MAINFILE);
 
 	boolean = g_key_file_get_boolean(config, "General",
-						"BackgroundScanning", &error);
+						CONF_BG_SCAN, &error);
 	if (error == NULL)
 		connman_settings.bg_scan = boolean;
 
 	g_clear_error(&error);
 
 	timeservers = g_key_file_get_string_list(config, "General",
-						"FallbackTimeservers", NULL, &error);
+					CONF_PREF_TIMESERVERS, NULL, &error);
 	if (error == NULL)
 		connman_settings.pref_timeservers = timeservers;
 
 	g_clear_error(&error);
 
 	str_list = g_key_file_get_string_list(config, "General",
-			"DefaultAutoConnectTechnologies", &len, &error);
+			CONF_AUTO_CONNECT, &len, &error);
 
 	if (error == NULL)
 		connman_settings.auto_connect =
@@ -190,7 +267,7 @@ static void parse_config(GKeyFile *config)
 	g_clear_error(&error);
 
 	str_list = g_key_file_get_string_list(config, "General",
-			"PreferredTechnologies", &len, &error);
+			CONF_PREFERRED_TECHS, &len, &error);
 
 	if (error == NULL)
 		connman_settings.preferred_techs =
@@ -201,7 +278,7 @@ static void parse_config(GKeyFile *config)
 	g_clear_error(&error);
 
 	str_list = g_key_file_get_string_list(config, "General",
-			"FallbackNameservers", &len, &error);
+			CONF_FALLBACK_NAMESERVERS, &len, &error);
 
 	if (error == NULL)
 		connman_settings.fallback_nameservers =
@@ -212,18 +289,57 @@ static void parse_config(GKeyFile *config)
 	g_clear_error(&error);
 
 	timeout = g_key_file_get_integer(config, "General",
-			"InputRequestTimeout", &error);
+			CONF_TIMEOUT_INPUTREQ, &error);
 	if (error == NULL && timeout >= 0)
 		connman_settings.timeout_inputreq = timeout * 1000;
 
 	g_clear_error(&error);
 
 	timeout = g_key_file_get_integer(config, "General",
-			"BrowserLaunchTimeout", &error);
+			CONF_TIMEOUT_BROWSERLAUNCH, &error);
 	if (error == NULL && timeout >= 0)
 		connman_settings.timeout_browserlaunch = timeout * 1000;
 
 	g_clear_error(&error);
+
+	interfaces = g_key_file_get_string_list(config, "General",
+			CONF_BLACKLISTED_INTERFACES, &len, &error);
+
+	if (error == NULL)
+		connman_settings.blacklisted_interfaces = interfaces;
+	else
+		connman_settings.blacklisted_interfaces =
+			g_strdupv(default_blacklist);
+
+	g_clear_error(&error);
+
+	boolean = g_key_file_get_boolean(config, "General",
+					CONF_ALLOW_HOSTNAME_UPDATES,
+					&error);
+	if (error == NULL)
+		connman_settings.allow_hostname_updates = boolean;
+
+	g_clear_error(&error);
+
+	boolean = g_key_file_get_boolean(config, "General",
+			CONF_SINGLE_TECH, &error);
+	if (error == NULL)
+		connman_settings.single_tech = boolean;
+
+	g_clear_error(&error);
+}
+
+static int config_init(const char *file)
+{
+	GKeyFile *config;
+
+	config = load_config(file);
+	check_config(config);
+	parse_config(config);
+	if (config != NULL)
+		g_key_file_free(config);
+
+	return 0;
 }
 
 static GMainLoop *main_loop = NULL;
@@ -305,6 +421,7 @@ static void disconnect_callback(DBusConnection *conn, void *user_data)
 	g_main_loop_quit(main_loop);
 }
 
+static gchar *option_config = NULL;
 static gchar *option_debug = NULL;
 static gchar *option_device = NULL;
 static gchar *option_plugin = NULL;
@@ -313,7 +430,7 @@ static gchar *option_noplugin = NULL;
 static gchar *option_wifi = NULL;
 static gboolean option_detach = TRUE;
 static gboolean option_dnsproxy = TRUE;
-static gboolean option_compat = FALSE;
+static gboolean option_backtrace = TRUE;
 static gboolean option_version = FALSE;
 
 static gboolean parse_debug(const char *key, const char *value,
@@ -328,6 +445,9 @@ static gboolean parse_debug(const char *key, const char *value,
 }
 
 static GOptionEntry options[] = {
+	{ "config", 'c', 0, G_OPTION_ARG_STRING, &option_config,
+				"Load the specified configuration file "
+				"instead of " CONFIGMAINFILE, "FILE" },
 	{ "debug", 'd', G_OPTION_FLAG_OPTIONAL_ARG,
 				G_OPTION_ARG_CALLBACK, parse_debug,
 				"Specify debug options to enable", "DEBUG" },
@@ -347,8 +467,9 @@ static GOptionEntry options[] = {
 	{ "nodnsproxy", 'r', G_OPTION_FLAG_REVERSE,
 				G_OPTION_ARG_NONE, &option_dnsproxy,
 				"Don't enable DNS Proxy" },
-	{ "compat", 'c', 0, G_OPTION_ARG_NONE, &option_compat,
-				"(obsolete)" },
+	{ "nobacktrace", 0, G_OPTION_FLAG_REVERSE,
+				G_OPTION_ARG_NONE, &option_backtrace,
+				"Don't print out backtrace information" },
 	{ "version", 'v', 0, G_OPTION_ARG_NONE, &option_version,
 				"Show version information and exit" },
 	{ NULL },
@@ -368,29 +489,38 @@ const char *connman_option_get_string(const char *key)
 
 connman_bool_t connman_setting_get_bool(const char *key)
 {
-	if (g_str_equal(key, "BackgroundScanning") == TRUE)
+	if (g_str_equal(key, CONF_BG_SCAN) == TRUE)
 		return connman_settings.bg_scan;
 
+	if (g_str_equal(key, CONF_ALLOW_HOSTNAME_UPDATES) == TRUE)
+		return connman_settings.allow_hostname_updates;
+
+	if (g_str_equal(key, CONF_SINGLE_TECH) == TRUE)
+		return connman_settings.single_tech;
+
 	return FALSE;
 }
 
 char **connman_setting_get_string_list(const char *key)
 {
-	if (g_str_equal(key, "FallbackTimeservers") == TRUE)
+	if (g_str_equal(key, CONF_PREF_TIMESERVERS) == TRUE)
 		return connman_settings.pref_timeservers;
 
-	if (g_str_equal(key, "FallbackNameservers") == TRUE)
+	if (g_str_equal(key, CONF_FALLBACK_NAMESERVERS) == TRUE)
 		return connman_settings.fallback_nameservers;
 
+	if (g_str_equal(key, CONF_BLACKLISTED_INTERFACES) == TRUE)
+		return connman_settings.blacklisted_interfaces;
+
 	return NULL;
 }
 
 unsigned int *connman_setting_get_uint_list(const char *key)
 {
-	if (g_str_equal(key, "DefaultAutoConnectTechnologies") == TRUE)
+	if (g_str_equal(key, CONF_AUTO_CONNECT) == TRUE)
 		return connman_settings.auto_connect;
 
-	if (g_str_equal(key, "PreferredTechnologies") == TRUE)
+	if (g_str_equal(key, CONF_PREFERRED_TECHS) == TRUE)
 		return connman_settings.preferred_techs;
 
 	return NULL;
@@ -410,7 +540,6 @@ int main(int argc, char *argv[])
 	GError *error = NULL;
 	DBusConnection *conn;
 	DBusError err;
-	GKeyFile *config;
 	guint signal;
 
 #ifdef NEED_THREADS
@@ -483,26 +612,28 @@ int main(int argc, char *argv[])
 
 	g_dbus_set_disconnect_function(conn, disconnect_callback, NULL, NULL);
 
-	__connman_log_init(argv[0], option_debug, option_detach);
+	__connman_log_init(argv[0], option_debug, option_detach,
+			option_backtrace, "Connection Manager", VERSION);
 
 	__connman_dbus_init(conn);
 
-	config = load_config(CONFIGDIR "/main.conf");
-	parse_config(config);
-	if (config != NULL)
-		g_key_file_free(config);
+	if (option_config == NULL)
+		config_init(CONFIGMAINFILE);
+	else
+		config_init(option_config);
 
-	__connman_storage_migrate();
+	__connman_inotify_init();
 	__connman_technology_init();
 	__connman_notifier_init();
+	__connman_agent_init();
 	__connman_service_init();
 	__connman_provider_init();
 	__connman_network_init();
 	__connman_device_init(option_device, option_nodevice);
 
-	__connman_agent_init();
 	__connman_ippool_init();
 	__connman_iptables_init();
+	__connman_firewall_init();
 	__connman_nat_init();
 	__connman_tethering_init();
 	__connman_counter_init();
@@ -530,6 +661,7 @@ int main(int argc, char *argv[])
 	__connman_wispr_init();
 	__connman_rfkill_init();
 
+	g_free(option_config);
 	g_free(option_device);
 	g_free(option_plugin);
 	g_free(option_nodevice);
@@ -544,8 +676,8 @@ int main(int argc, char *argv[])
 	__connman_wpad_cleanup();
 	__connman_dhcpv6_cleanup();
 	__connman_dhcp_cleanup();
-	__connman_provider_cleanup();
 	__connman_plugin_cleanup();
+	__connman_provider_cleanup();
 	__connman_connection_cleanup();
 	__connman_timeserver_cleanup();
 	__connman_session_cleanup();
@@ -560,21 +692,23 @@ int main(int argc, char *argv[])
 	__connman_config_cleanup();
 	__connman_manager_cleanup();
 	__connman_counter_cleanup();
-	__connman_agent_cleanup();
 	__connman_tethering_cleanup();
 	__connman_nat_cleanup();
+	__connman_firewall_cleanup();
 	__connman_iptables_cleanup();
 	__connman_ippool_cleanup();
 	__connman_device_cleanup();
 	__connman_network_cleanup();
 	__connman_service_cleanup();
+	__connman_agent_cleanup();
 	__connman_ipconfig_cleanup();
 	__connman_notifier_cleanup();
 	__connman_technology_cleanup();
+	__connman_inotify_cleanup();
 
 	__connman_dbus_cleanup();
 
-	__connman_log_cleanup();
+	__connman_log_cleanup(option_backtrace);
 
 	dbus_connection_unref(conn);
 
@@ -586,6 +720,7 @@ int main(int argc, char *argv[])
 	g_free(connman_settings.auto_connect);
 	g_free(connman_settings.preferred_techs);
 	g_strfreev(connman_settings.fallback_nameservers);
+	g_strfreev(connman_settings.blacklisted_interfaces);
 
 	g_free(option_debug);
 
